import random
from graphics_and_games_klassen import Figur
from helferklassen import *
from intern.zeichenfenster import Zeichenfenster
import pygame

class LabyrinthElement:
    """Oberklasse aller Bausteine des Labyrinths"""
    seitenlaenge = 20

    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.darstellung = Rechteck()
        self.darstellung.PositionSetzen(x * self.seitenlaenge, y * self.seitenlaenge)
        self.darstellung.GroesseSetzen(self.seitenlaenge, self.seitenlaenge)

    def XPositionGeben(self):
        """Gibt die x-Position des Elements zurück."""
        return self.x

    def YPositionGeben(self):
        """Gibt die y-Position des Elements zurück."""
        return self.y

    def SichtbarkeitSetzen(self, sichtbar):
        """Setzt die Sichtbarkeit des Elements."""
        self.darstellung.SichtbarkeitSetzen(sichtbar)


class Mauer(LabyrinthElement):
    """Ein Mauerbaustein des Labyrinths"""
    def __init__(self, x, y):
        super().__init__(x, y)
        self.darstellung.FarbeSetzen("schwarz")


class Gang(LabyrinthElement):
    """Ein Gangbaustein des Labyrinths"""
    def __init__(self, x, y):
        super().__init__(x, y)
        self.darstellung.FarbeSetzen("weiss")


class Ausgang(Gang):
    """Der Ausgangbaustein des Labyrinths"""
    def __init__(self, x, y):
        super().__init__(x, y)
        self.darstellung.FarbeSetzen("gruen")

class Knoten:
    """Beschreibt einen Knoten"""
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.symbol = KnotenSymbol(x * LabyrinthElement.seitenlaenge + LabyrinthElement.seitenlaenge // 2,
                                   y * LabyrinthElement.seitenlaenge + LabyrinthElement.seitenlaenge // 2,
                                   (LabyrinthElement.seitenlaenge - 4) // 2, "weiss", "")
        self.vorgaenger = None

    def XGeben(self):
        """Meldet die x-Koordinate des Knotens zurück."""
        return self.x

    def YGeben(self):
        """Meldet die y-Koordinate des Knotens zurück."""
        return self.y

    def SymbolGeben(self):
        """Meldet das Symbol des Knotens zurück."""
        return self.symbol

    def VorgaengerGeben(self):
        """Meldet den Vorgänger des Knotens zurück."""
        return self.vorgaenger

    def VorgaengerSetzen(self, v):
        """Setzt den Vorgänger für den Weg zum Ausgang."""
        self.vorgaenger = v

    def FarbeSetzen(self, f):
        """Setzt die Farbe der Darstellung."""
        self.symbol.FarbeSetzen(f)


class GraphMatrix:
    """Verwaltet einen ungerichteten, gewichteten Graphen mittels Adjazenzmatrix"""
    def __init__(self):
        self.knoten = []
        self.matrix = []
        self.kanten = []
        self.besuchteKnoten = []
        self.weg = []

    def KnotenEinfuegen(self, x, y):
        """Einfügen eines neuen Knoten in den Graphen."""
        self.knoten.append(Knoten(x, y))
        for row in self.matrix:
            row.append(-1)
        self.matrix.append([-1] * (len(self.knoten)))

    def KnotenGeben(self, x, y):
        """Gibt den Knoten mit den angegebenen Koordinaten zurück."""
        for k in self.knoten:
            if k.XGeben() == x and k.YGeben() == y:
                return k
        return None

    def KnotenNummerGeben(self, x, y):
        """Gibt die interne Nummer des Knotens."""
        for i, k in enumerate(self.knoten):
            if k.XGeben() == x and k.YGeben() == y:
                return i
        return -1

    def KanteEinfuegen(self, x1, y1, x2, y2):
        """Einfügen einer Kante in den Graphen."""
        von_nummer = self.KnotenNummerGeben(x1, y1)
        nach_nummer = self.KnotenNummerGeben(x2, y2)
        if von_nummer != -1 and nach_nummer != -1 and von_nummer != nach_nummer:
            self.matrix[von_nummer][nach_nummer] = 1
            self.matrix[nach_nummer][von_nummer] = 1
            self.kanten.append(KantenSymbol(self.knoten[von_nummer].SymbolGeben(), self.knoten[nach_nummer].SymbolGeben(), False, "", 3, "blau"))

    def KnotenAnzahlGeben(self):
        """Gibt die Anzahl der Knoten des Graphen zurück."""
        return len(self.knoten)

    def KanteGewichtGeben(self, x1, y1, x2, y2):
        """Gibt die Gewichtung einer Kante zurück."""
        von_nummer = self.KnotenNummerGeben(x1, y1)
        nach_nummer = self.KnotenNummerGeben(x2, y2)
        if von_nummer != -1 and nach_nummer != -1:
            return self.matrix[von_nummer][nach_nummer]
        return -1

    def ZurueckSetzen(self):
        """Löscht die Kanten und Knoten des Graphen. Die Anzeige wird auch gelöscht."""
        for k in self.kanten:
            k.Entfernen()
        for k in self.knoten:
            k.SymbolGeben().Entfernen()
        self.knoten.clear()
        self.kanten.clear()
        self.matrix = []

    def SichtbarkeitSetzen(self, sichtbar):
        """Setzt die Sichtbarkeit des Graphen."""
        for k in self.kanten:
            k.SichtbarkeitSetzen(sichtbar)
        for k in self.knoten:
            k.SymbolGeben().SichtbarkeitSetzen(sichtbar)
            
    def TiefensucheAusfuehren(self, x1, y1, x2, y2):
        """
        Führt die Tiefensuche vom Startknoten aus durch.
        :param x1: x-Koordinate des Startknotens
        :param y1: y-Koordinate des Startknotens
        :param x2: x-Koordinate des Zielknotens
        :param y2: y-Koordinate des Zielknotens
        :return: Weg als Liste von Knoten
        """
        self.besuchteKnoten.clear()
        self.weg.clear()
        self.KnotenBesuchen(self.KnotenNummerGeben(x1, y1), self.KnotenNummerGeben(x2, y2))
        return self.weg
    
    def KnotenBesuchen(self, aktuell, ziel):
        """
        Setzt den Rekursionsschritt bei einem Knoten um.
        Prüft, ob ein geeigneter Knoten gefunden wurde.

        :param aktuell: Die Nummer des aktuell zu besuchenden Knotens
        :param ziel: Die Nummer des Zielknotens
        :return: True, wenn ein geeigneter Knoten gefunden wurde
        """
        # Den Rumpf gemäß Aufgabe ergänzen
        pass


class Labyrinth:
    """Verwaltet das Labyrinth"""
    def __init__(self, breite, hoehe):
        self.zufall = random.Random()
        self.elemente = [[None for _ in range(hoehe)] for _ in range(breite)]
        self.xAusgang = -1
        self.yAusgang = -1
        self.xStart = -1
        self.yStart = -1

    def Generieren(self):
        """Generiert das Labyrinth"""
        dx, dy = 0, 0
        xmax = len(self.elemente)
        ymax = len(self.elemente[0])
        minSchritte = xmax * ymax // 3
        schritte = 1
        x = random.randint(1, xmax - 2)
        y = random.randint(1, ymax - 2)
        self.elemente[x][y] = Gang(x, y)

        while True:
            while True:
                dx = random.randint(-1, 1)
                dy = random.randint(-1, 1)
                if abs(dx) != abs(dy):
                    break
            
            x += dx
            y += dy
            if (x == 0) or (y == 0) or (x == xmax - 1) or (y == ymax - 1):
                if schritte > minSchritte and self.elemente[x][y] is None:
                    self.elemente[x][y] = Ausgang(x, y)
                    self.xAusgang = x
                    self.yAusgang = y
                    break
                else:
                    x -= dx
                    y -= dy
            elif self.elemente[x][y] is None:
                if (self.elemente[x - 1][y - 1] is not None and self.elemente[x - 1][y] is not None and self.elemente[x][y - 1] is not None or
                    self.elemente[x + 1][y - 1] is not None and self.elemente[x + 1][y] is not None and self.elemente[x][y - 1] is not None or
                    self.elemente[x + 1][y + 1] is not None and self.elemente[x + 1][y] is not None and self.elemente[x][y + 1] is not None or
                    self.elemente[x - 1][y + 1] is not None and self.elemente[x - 1][y] is not None and self.elemente[x][y + 1] is not None):
                    x -= dx
                    y -= dy
                else:
                    schritte += 1
                    self.elemente[x][y] = Gang(x, y)

        for b in range(len(self.elemente)):
            for h in range(len(self.elemente[b])):
                if self.elemente[b][h] is None:
                    self.elemente[b][h] = Mauer(b, h)


    def GraphErzeugen(self, graph):
        """
        Erzeugt die Graphendarstellung.
        """
        for x in range(len(self.elemente)):
            for y in range(len(self.elemente[x])):
                if not isinstance(self.elemente[x][y], Mauer):
                    graph.KnotenEinfuegen(x, y)
                    if isinstance(self.elemente[x][y], Ausgang):
                        graph.KnotenGeben(x, y).FarbeSetzen("gruen")
                    if x > 0 and not isinstance(self.elemente[x - 1][y], Mauer):
                        graph.KanteEinfuegen(x, y, x - 1, y)
                    if y > 0 and not isinstance(self.elemente[x][y - 1], Mauer):
                        graph.KanteEinfuegen(x, y, x, y - 1)

    def SichtbarkeitSetzen(self, sichtbar):
        """Setzt die Sichtbarkeit des Labyrinths."""
        for x in range(len(self.elemente)):
            for y in range(len(self.elemente[x])):
                if self.elemente[x][y] is not None:
                    self.elemente[x][y].SichtbarkeitSetzen(sichtbar)

    def XAusgangGeben(self):
        """Meldet die x-Koordinate des Ausgangs zurück."""
        return self.xAusgang

    def YAusgangGeben(self):
        """Meldet die y-Koordinate des Ausgangs zurück."""
        return self.yAusgang

    def StartpunktGenerieren(self):
        """Ermittelt den Startpunkt."""
        while True:
            x1 = random.randint(1, len(self.elemente) - 2)
            y1 = random.randint(1, len(self.elemente[0]) - 2)
            if not isinstance(self.elemente[x1][y1], Mauer):
                break

        while True:
            x2 = random.randint(1, len(self.elemente) - 2)
            y2 = random.randint(1, len(self.elemente[0]) - 2)
            if not isinstance(self.elemente[x2][y2], Mauer):
                break

        while True:
            x3 = random.randint(1, len(self.elemente) - 2)
            y3 = random.randint(1, len(self.elemente[0]) - 2)
            if not isinstance(self.elemente[x3][y3], Mauer):
                break
        if abs(x1 - self.xAusgang) + abs(y1 - self.yAusgang) > abs(x2 - self.xAusgang) + abs(y2 - self.yAusgang):
            if abs(x1 - self.xAusgang) + abs(y1 - self.yAusgang) > abs(x3 - self.xAusgang) + abs(y3 - self.yAusgang):
                self.xStart = x1
                self.yStart = y1
            elif abs(x2 - self.xAusgang) + abs(y2 - self.yAusgang) > abs(x3 - self.xAusgang) + abs(y3 - self.yAusgang):
                self.xStart = x2
                self.yStart = y2
            else:
                self.xStart = x3
                self.yStart = y3
        elif abs(x2 - self.xAusgang) + abs(y2 - self.yAusgang) > abs(x3 - self.xAusgang) + abs(y3 - self.yAusgang):
            self.xStart = x2
            self.yStart = y2
        else:
            self.xStart = x3
            self.yStart = y3
            
    def XStartGeben(self):
        """
        Meldet die x-Koordinate des Ausgangs zurück.
        :return: x-Koordinate des Ausgangs
        """
        return self.xStart

    def YStartGeben(self):
        """
        Meldet die y-Koordinate des Ausgangs zurück.
        :return: y-Koordinate des Ausgangs
        """
        return self.yStart

class Theseus(Figur):
    """Theseus erhält einen Weg, auf dem er zum Ausgang findet"""
    
    def __init__(self, x, y):
        super().__init__()
        self.x1 = x
        self.y1 = y
        self.weg = None
        self.schritte = 0
        self.seitenlänge = 20
        self.GroesseSetzen(self.seitenlänge)
        self.FigurFestlegen()
        self.PositionSetzen(
            x * self.seitenlänge + self.seitenlänge // 2, y  * self.seitenlänge + self.seitenlänge // 2)

    def FigurFestlegen(self):
        """Legt das Aussehen von Theseus fest."""
        self.EigeneFigurLoeschen()
        self.FigurteilFestlegenRechteck(-35, 0, 70, 25, "rot")
        self.FigurteilFestlegenEllipse(-25, 25, 10, 10, "schwarz")
        self.FigurteilFestlegenEllipse(15, 25, 10, 10, "schwarz")
        self.FigurteilFestlegenEllipse(-20, -40, 40, 40, "rot")
        self.FigurteilFestlegenEllipse(5, -27, 15, 15, "gelb")


    def WegSetzen(self, weg):
        """Legt den Weg zum Ausgang fest."""
        self.weg = weg
        self.schritte = 1

    def AktionAusfuehren(self):
        """Bewegt Theseus ein Feld in Richtung Ausgang."""
        if self.weg is not None and self.schritte < len(self.weg):
            k = self.weg[self.schritte]
            if k.XGeben() == self.x1 + 1:
                self.WinkelSetzen(0)
            elif k.XGeben() == self.x1 - 1:
                self.WinkelSetzen(180)
            elif k.YGeben() == self.y1 - 1:
                self.WinkelSetzen(90)
            elif k.YGeben() == self.y1 + 1:
                self.WinkelSetzen(270)
            
            self.Gehen(self.seitenlänge)
            self.x1 = k.XGeben()
            self.y1 = k.YGeben()
            self.schritte += 1
        
class Rahmen:
    """Rahmenklasse für die Wegesuche aus dem Labyrinth."""
    def __init__(self):
        self.l = Labyrinth(20, 13)
        self.l.Generieren()
        self.g = GraphMatrix()
        self.l.GraphErzeugen(self.g)
        self.t = None

    def GraphAnzeigen(self):
        """Zeigt nur den Graphen an."""
        self.l.SichtbarkeitSetzen(False)
        self.g.SichtbarkeitSetzen(True)

    def LabyrinthAnzeigen(self):
        """Zeigt nur das Labyrinth an."""
        self.l.SichtbarkeitSetzen(True)
        self.g.SichtbarkeitSetzen(False)

    def TheseusPositionieren(self):
        """Positioniert den verirrten Theseus auf einen geeigenten Startpunkt."""
        if self.t:
            self.t.Entfernen()
        self.LabyrinthAnzeigen()
        self.l.StartpunktGenerieren()
        self.t = Theseus(self.l.XStartGeben(), self.l.YStartGeben())

    def FluchtwegBerechnen(self):
        """Ermittelt den Weg aus dem Labyrinth und bereitet den Gang zum Ausgang vor."""
        if self.t:
            self.t.WegSetzen(self.g.TiefensucheAusfuehren(self.l.XStartGeben(), self.l.YStartGeben(), self.l.XAusgangGeben(), self.l.YAusgangGeben()))

    def FluchtAusfuehren(self):
        """Setzt Theseus auf eine Startposition und setzt den Weg zum Ausgang."""
        self.TheseusPositionieren()
        self.FluchtwegBerechnen()
        


r = Rahmen()
r.FluchtAusfuehren()


