from helferklassen import Knoten, KantenSymbol, Lesen
from graphics_and_games_klassen import Ereignisbehandlung
import time

class GraphMatrix(Ereignisbehandlung):

    def __init__(self):
        """
        Baut die Datenstruktur auf
        """
        super().__init__()
        self.knoten = []
        self.matrix = []
        self.kanten = []
        self.besuchteKnoten = []
        self.aktuellerWeg = []
        self.optimalerWeg = []
        self.optimaleLaenge = float('inf')
        
        
    def KnotenEinfuegen(self, bezeichner, x, y):
        """
        Einfügen eines neuen Knoten in den Graphen
        """
        self.knoten.append(Knoten(bezeichner, x, y))
        for zeile in self.matrix:
            zeile.append(-1)
        self.matrix.append([])
        for zeile in self.matrix:
            self.matrix[-1].append(-1)

    def KnotenGeben(self, bezeichner):
        """
        Gibt den Knoten mit dem angegebenen Bezeichner Zurueck.
        """
        for k in self.knoten:
            if k.BezeichnerGeben() == bezeichner:
                return k
        return None
 
    def KnotenNummerGeben(self, bezeichner):
        """
        Gibt die interne Nummer des Knotens
        """
        for i in range(len(self.knoten)):
            if self.knoten[i].BezeichnerGeben() == bezeichner:
                return i
        return -1

    def KnotenBezeichnerGeben(self, knotenNummer):
        """
        Gibt die Bezeichnung eines Knotens mit der internen Knotennummer
        """
        if knotenNummer < len(self.knoten) and knotenNummer >= 0:
            return self.knoten[knotenNummer].BezeichnerGeben()
        else:
            return ""

    def KanteEinfuegen(self, von, nach, gewichtung):
        """
        Einfügen einer Kante in den Graphen
        """
        vonNummer = self.KnotenNummerGeben(von)
        nachNummer = self.KnotenNummerGeben(nach)
        if vonNummer != -1 and nachNummer != -1 and vonNummer != nachNummer:
            self.matrix[vonNummer][nachNummer] = gewichtung
            self.matrix[nachNummer][vonNummer] = gewichtung
            self.kanten.append(KantenSymbol(self.knoten[vonNummer].SymbolGeben(), self.knoten[nachNummer].SymbolGeben(), False, str(gewichtung), 3, "blau"))

    def Ausgeben(self):
        """
        Gibt die Adjazenzmatrix des Graphen in der Konsole aus
        """
        # Kopfzeile
        print("    ", end='')
        for k in self.knoten:
            print((k.BezeichnerGeben() + "    ")[0:4], end='')
        print("")

        for i in range(len(self.knoten)):
            print((self.knoten[i].BezeichnerGeben() + "    ")[0:4], end='')
            for j in range(len(self.knoten)):
                if self.matrix[i][j] != -1:
                    print((str(self.matrix[i][j]) + "   ")[0:4], end='')
                else:
                    print("    ", end='')
            print("")

    def KnotenAnzahlGeben(self):
        """
        Gibt die Anzahl der Knoten des Graphen Zurueck
        """
        return len(self.knoten)

    def KanteGewichtGeben(self, von, nach):
        """
        Gibt die Gewichtung einer Kante Zurueck
        Die Kante ist durch einen Anfangsknoten und einen Endknoten festgelegt
        """
        vonNummer = self.KnotenNummerGeben(von)
        nachNummer = self.KnotenNummerGeben(nach)
        if vonNummer != -1 and nachNummer!=-1:
            return self.matrix[vonNummer][nachNummer]
        else:
            return -1


    def ZurueckSetzen(self):
        """Löscht die Kanten und Knoten des Graphen. Die Anzeige wird auch gelöscht."""
        for k in self.kanten:
            k.Entfernen()
        for k in self.knoten:
            k.SymbolGeben().Entfernen()
        self.knoten.clear()
        self.kanten.clear()
        self.matrix.clear()


    def AlleVerbindungenBerechnen(self):
        """
        Berechnet alle Knotenverbindungen nach Floyd-Warshall.
        """
        n = len(self.matrix)
        for k in range(n):
            for i in range(n):
                for j in range(n):
                    if i != j and self.matrix[i][k] != -1 and self.matrix[k][j] != -1:
                        if self.matrix[i][j] == -1 or self.matrix[i][j] > self.matrix[i][k] + self.matrix[k][j]:
                            self.matrix[i][j] = self.matrix[i][k] + self.matrix[k][j]

    def KnotenBesuchen(self, aktuell, zielKnoten, länge):
        """
        Setzt den Rekursionsschritt bei einem Knoten um.
        """
        # Rekursion laut Lehrtext hier ergänzen
        pass

    def TiefensucheAusfuehren(self, startKnoten):
        """
        Führt die Tiefensuche vom Startknoten aus durch.
        """
        # Vorbereitung hier ergänzen
        print("Der Weg hat die Länge: ", self.optimaleLänge)
        print("Er führt über")
        for i in self.optimalerWeg:
            print(self.knoten[i].BezeichnerGeben())


class Beispiele:
    """Beispiele für diverse Graphen."""

    def __init__(self):
        """Legt das Graphenobjekt an."""
        self.g = GraphMatrix()
        self.l = Lesen()

    def StandardbeispielLaden(self):
        """
        Lädt das Standardbeispiel.
        """
        self.g.ZurueckSetzen()
        self.l.LesenDatenbank("Autobahn.grdb", self.g)
        self.g.AlleVerbindungenBerechnen()

    def KleinesBeispielLaden(self):
        """
        Lädt das kleine Beispiel.
        """
        self.g.ZurueckSetzen()
        self.l.LesenDatenbank("Autobahn_klein.grdb", self.g)
        self.g.AlleVerbindungenBerechnen()

    def TiefensucheAusfuehren(self, nameStartKnoten):
        """
        Führt die Tiefensuche vom Startknoten zurück zum Startknoten durch.
        :param nameStartKnoten: der Name des Startknotens
        """
        if self.g is not None:
            start_knoten_index = self.g.KnotenNummerGeben(nameStartKnoten)
            self.g.TiefensucheAusfuehren(start_knoten_index)
            
            
# Anwendung
b = Beispiele()
b.KleinesBeispielLaden()
b.TiefensucheAusfuehren("M")