from helferklassen import KnotenSymbol, KantenSymbol, Lesen

class Knoten:
    def __init__(self, bezeichner, x, y):
        """
        Besetzt die Attribute und legt das Knotensymbol an.
        """
        self.bezeichner = bezeichner
        self.symbol = KnotenSymbol(x, y, 20, "weiss", bezeichner)
        self.adjazenzliste = []

    def BezeichnerGeben(self):
        """
        Meldet den Bezeichner des Knotens zurück.
        """
        return self.bezeichner

    def SymbolGeben(self):
        """
        Meldet das Symbol des Knotens zurück.
        """
        return self.symbol

    def FarbeSetzen(self, f):
        """
        Setzt die Farbe der Darstellung.
        """
        self.symbol.FarbeSetzen(f)

    def AdjazenzlisteGeben(self):
        """
        Meldet die Adjazenzliste des Knotens zurück.
        """
        return self.adjazenzliste

class Kante:
    def __init__(self, gewicht, ziel):
        """
        Attribute laut Beschreibung ergänzen.
        """
        self.gewicht = gewicht
        self.ziel = ziel

    def GewichtGeben(self):
        """
        Meldet das Gewicht der Kante zurück.
        """
        return self.gewicht

    def ZielGeben(self):
        """
        Meldet den Zielknoten der Kante zurück.
        """
        return self.ziel

class GraphListe:
    def __init__(self):
        """
        Baut die Datenstruktur auf.
        """
        self.knoten = []
        self.kanten = []
        self.besuchteKnoten = []

    def KnotenEinfuegen(self, bezeichner, x, y):
        """
        Einfügen eines neuen Knoten in den Graphen.
        """
        self.knoten.append(Knoten(bezeichner, x, y))

    def KnotenGeben(self, bezeichner):
        """
        Gibt den Knoten mit dem angegebenen Bezeichner zurück.
        """
        for k in self.knoten:
            if k.BezeichnerGeben() == bezeichner:
                return k
        return None

    def KnotenNummerGeben(self, bezeichner):
        """
        Gibt die interne Nummer des Knotens.
        """
        for i in range(len(self.knoten)):
            if self.knoten[i].BezeichnerGeben() == bezeichner:
                return i
        return -1

    def KnotenBezeichnerGeben(self, knotenNummer):
        """
        Gibt die Bezeichnung eines Knotens mit der internen Knotennummer.
        """
        if 0 <= knotenNummer < len(self.knoten):
            return self.knoten[knotenNummer].BezeichnerGeben()
        return ""

    def KanteEinfuegen(self, von, nach, gewichtung):
        """
        Einfügen einer Kante in den Graphen.
        """
        vonNummer = self.KnotenNummerGeben(von)
        nachNummer = self.KnotenNummerGeben(nach)
        if vonNummer != -1 and nachNummer != -1 and vonNummer != nachNummer:
            self.knoten[vonNummer].AdjazenzlisteGeben().append(Kante(gewichtung, self.knoten[nachNummer]))
            self.knoten[nachNummer].AdjazenzlisteGeben().append(Kante(gewichtung, self.knoten[vonNummer]))
            self.kanten.append(KantenSymbol(self.knoten[vonNummer].SymbolGeben(), self.knoten[nachNummer].SymbolGeben(), False, str(gewichtung), 3, "blau"))

    def Ausgeben(self):
        """
        Gibt die Adjazenzmatrix des Graphen in der Konsole aus.
        """
        for k in self.knoten:
            print(f"{k.BezeichnerGeben()}: ", end="")
            for ka in k.AdjazenzlisteGeben():
                print(f"({ka.GewichtGeben()}-> {ka.ZielGeben().BezeichnerGeben()}) ", end="")
            print("")

    def KnotenAnzahlGeben(self):
        """
        Gibt die Anzahl der Knoten des Graphen zurück.
        """
        return len(self.knoten)

    def KanteGewichtGeben(self, von, nach):
        """
        Gibt die Gewichtung einer Kante zurück.
        """
        vonNummer = self.KnotenNummerGeben(von)
        nachKnoten = self.KnotenGeben(nach)
        if vonNummer != -1 and nachKnoten:
            for k in self.knoten[vonNummer].AdjazenzlisteGeben():
                if k.ZielGeben() == nachKnoten:
                    return k.GewichtGeben()
        return -1

    def ZurueckSetzen(self):
        """
        Löscht die Kanten und Knoten des Graphen. Die Anzeige wird auch gelöscht.
        """
        for k in self.kanten:
            k.Entfernen()
        for k in self.knoten:
            k.SymbolGeben().Entfernen()
        self.knoten.clear()
        self.kanten.clear()

    def KnotenBesuchen(self, aktuell):
        """
        Setzt den Rekursionsschritt bei einem Knoten um.
        :param aktuell: der aktuelle Knoten
        """
        self.besuchteKnoten.append(aktuell)
        for kante in aktuell.AdjazenzlisteGeben():
            if kante.ZielGeben() not in self.besuchteKnoten:
                self.KnotenBesuchen(kante.ZielGeben())
        aktuell.FarbeSetzen("grün")
        
    def TiefensucheAusführen(self, startKnoten):
        """
        Führt die Tiefensuche vom Startknoten aus durch.
        :param startKnoten: der Startknoten
        """
        self.besuchteKnoten.clear()
        self.KnotenBesuchen(startKnoten)


class Beispiele:
    """Beispiele für diverse Graphen."""

    def __init__(self):
        """Legt das Graphenobjekt an."""
        self.g = GraphListe()
        self.l = Lesen()

    def AusfuehrenAbiturfahrt(self):
        """Legt einen Graphen für eine Abiturfahrt an."""
        self.g.ZurueckSetzen()
        if self.l.LesenDatenbank("Abiturfahrt.grdb", self.g):
            pass  # g.Ausgeben()

    def AusfuehrenAutobahn(self):
        """Legt einen Graphen für den süddeutschen Teil des Autobahnnetzes an."""
        self.g.ZurueckSetzen()
        if self.l.LesenDatenbank("Autobahn.grdb", self.g):
            pass  # g.Ausgeben()

    def AusfuehrenICE(self):
        """Legt einen Graphen für das deutsche ICE-Netz an."""
        self.g.ZurueckSetzen()
        if self.l.LesenDatenbank("ICENetz.grdb", self.g):
            pass  # g.Ausgeben()

    def AusfuehrenBeispielgraph(self):
        """Legt einen zusammenhängenden Graphen an."""
        self.g.ZurueckSetzen()
        if self.l.LesenDatenbank("Beispielgraph.grdb", self.g):
            pass  # g.Ausgeben()

    def AusfuehrenBeispielgraph2(self):
        """Legt einen nicht zusammenhängenden Graphen an."""
        self.g.ZurueckSetzen()
        if self.l.LesenDatenbank("Beispielgraph2.grdb", self.g):
            pass  # g.Ausgeben()

    def TiefensucheAusfuehren(self, nameStartKnoten):
        """Führt die Tiefensuche vom Startknoten aus durch."""
        if self.g:
            self.g.TiefensucheAusführen(self.g.KnotenGeben(nameStartKnoten))

    def ZusammenhangTesten(self):
        """Testet, ob der Graph zusammenhängend ist."""
        if self.g.ZusammenhangTesten():
            print("Der Graph ist zusammenhängend.")
        else:
            print("Der Graph ist nicht zusammenhängend.")

# Anwendung
b = Beispiele()
b.AusfuehrenBeispielgraph2()
b.TiefensucheAusfuehren("A")
